module RefacRedunDec (convertPatsToPn, convertToPat, refacRedunDec, removeRedun, removeRedunWhere) where
 

import AbstractIO
import Maybe
import List
import RefacUtils

data Patt = Match HsMatchP | MyPat HsDeclP | Def [Char]

refacRedunDec args
  = do let fileName   = args!!0
           beginRow   = read (args!!1)::Int
           beginCol   = read (args!!2)::Int
           endRow     = read (args!!3)::Int
           endCol     = read (args!!4)::Int
       AbstractIO.putStrLn "refacRedunDec"

       -- Parse the input file.
       modInfo@(inscps, exps, mod, tokList) <- parseSourceFile fileName

       -- Find the function that's been highlighted as the refactree

       let (loc, pnt, pats, exp, wh, p)
             = findDefNameAndExp tokList
                                 (beginRow, beginCol)
                                 (endRow, endCol)
                                 mod 
       (wh', newExp) <- doRefac wh exp
     
       (mod',((tokList',modified),_))<-doRemovingWhere fileName mod tokList exp newExp wh'
     
--       AbstractIO.putStrLn (show mod')
       
       (mod'',((tokList'',modified'),_))<-doRemoving fileName mod' tokList' exp newExp wh'
     
   --    AbstractIO.putStrLn (show mod'')
       
       writeRefactoredFiles False [((fileName,modified'), (tokList'',mod''))]
       
    --   error (show mod')
       AbstractIO.putStrLn "Completed.\n"

doRefac wh exp = do
                 (decls, newExp) <- removeRedun wh exp
                 (newWhere) <- removeRedunWhere wh newExp
                 return (newWhere, newExp)

convertToPat [] _ = []
convertToPat _ [] = []
convertToPat (pat@(Pat (HsPId (HsVar (PNT (PN x y) Value s)))):ps) ((PN v z):xs)
      | x == v && y == z = pat : (convertToPat ps xs)
      | otherwise        = convertToPat ps ((PN v z):xs)


convertPatsToPn [] = []
convertPatsToPn ((Pat (HsPId (HsVar (PNT (PN x y)Value s )))):ps)
                           = (PN x y): (convertPatsToPn ps)
convertPatsToPn _ = []


--removeRedun :: [HsDeclP] -> HsExpP -> m ([HsDeclP], HsExpP)

removeRedun wh (Exp(HsLet ds e)) = do
                                   (free, dec) <- hsFreeAndDeclaredPNs e
                                   let decls = checkFree free ds
                                   let decls' = checkFree free wh
                                   newWhere <- checkWhere wh decls
                                   moreDecls <- checkRestDecls decls' wh
                                   if decls /= [] 
                                      then do                                   
                                        return ((nub (moreDecls++(decls'++newWhere))), (Exp(HsLet decls e)))
                                      else do
                                        return ((nub (moreDecls++(decls'++newWhere))), e)
                                      where
                                         checkWhere wh [] = do return wh
                                         checkWhere wh ((Dec(HsPatBind _ _ (HsBody e)_)):xs) 
                                                      = do
                                                           (free, dec) <- hsFreeAndDeclaredPNs e 
                                                           let decls = checkFree free wh
                                                           newDecls <- checkWhere decls xs
                                                           moreDecls <- checkRestDecls newDecls wh
                                                           return (nub (moreDecls ++ newDecls))
removeRedun wh (exp@(_)) = do return (wh, exp)
removeRedunWhere wh (exp@(_)) = do
                      (free, dec) <- hsFreeAndDeclaredPNs exp
                      let decls = checkFree free wh
                      -- check each declaration that is left for 
                      -- declarations that are still needed
                      moreDecls <- checkRestDecls decls wh
                      return (nub (decls++moreDecls))

checkRestDecls [] _ = return []                         
checkRestDecls wh [] = return []
checkRestDecls ((Dec (HsPatBind loc1 ps (HsBody e) ds)):rest) (x:[])
                      = do
                          (free, dec) <- hsFreeAndDeclaredPNs e
                          let decls = checkFree free [x]
                          return decls
 
checkRestDecls ((Dec (HsPatBind loc1 ps (HsBody e) ds)):rest) (x:xs) 
                      = do
                          (free, dec) <- hsFreeAndDeclaredPNs e
                          let decls = checkFree free (x:xs)
                         -- moreDecls <- checkRestDecls [x] xs
                          moreDecls <- checkRestDecls decls (x:xs)
                          moreDecls' <- checkRestDecls rest (x:xs)
                          return (decls ++ (moreDecls ++ moreDecls'))

-- Stafunski function to traverse the AST and find the occurance of the exp that we
-- are interested in. When the exp is found, the redundant declarations are calculated
-- and rmDecls in then called. 
-- the rusult is then a new expression with the new list of declarations returned by
-- rmDecls.
doRemoving fileName mod tokList exp newExp wh
    =  runStateT (applyTP (stop_tdTP (failTP `adhocTP` (rmInLet newExp))) mod) ((tokList,unmodified),fileName)
                       
           where                 
             --1. The definition to be removed is a local declaration in a let expression
             rmInLet (new@(Exp (HsLet decs e'))::HsExpP) (letExp@(Exp (HsLet ds e))::HsExpP) 
               | sameOccurrence letExp exp == True
                  =do
                      let redunDecs = ds \\ decs
                   --   error (show redunDecs)
                      ds'<- rmDecls ds [] redunDecs
                      if ds'==[] then return e'
                                 else return (Exp (HsLet ds' e))
               | otherwise = error ((show letExp) ++ "\n" ++ (show exp))

             
             rmInLet e1 (Exp (HsLet ds e)::HsExpP) 
               | sameOccurrence e1 e == True
                  = do
                      ds' <- rmDecls ds [] ds
                      if ds' == [] then return e1
                                   else return (Exp (HsLet ds' e))
               | otherwise
                   = return e
             rmInLet e1 e2 = fail ""
             rmDecls wh d (((Dec (HsPatBind loc1 pn (HsBody rhs) ds))::HsDeclP):xs)
                  = do
                      let newName = pNTtoPN (patToPNT pn)
                      decls' <- rmDecl newName False wh
                      decls'' <- rmDecls wh decls' xs
                      return decls''
             rmDecls _ d _ = return d

         
doRemovingWhere fileName mod tokList exp newExp wh
     =  runStateT (applyTP (stop_tdTP (failTP `adhocTP` (rmInExp exp))) mod) ((tokList,unmodified),fileName)
          where
             rmInExp exp (pat@(Dec (HsPatBind loc p (HsBody rhs) ds))::HsDeclP)
                | sameOccurrence exp rhs 
                   =do
                       let redunDecs = ds \\ wh
                       ds'<- rmDecls ds [] redunDecs                      
                  --     error (show ds')
                       if ds' == [] then fail ""
                                    else
                                      return (Dec (HsPatBind  loc p (HsBody rhs) ds'))                    
             rmInExp _ e2 = fail ""
             --error ((show e1) ++ " " ++ (show e2)  )
             rmDecls wh d (((Dec (HsPatBind loc1 pn (HsBody rhs) ds))::HsDeclP):xs)
                  = do
                      let newName = pNTtoPN (patToPNT pn)
                      decls' <- rmDecl newName False wh
                      decls'' <- rmDecls wh decls' xs
                      return decls''
             rmDecls _ d _ = return d

newMatch loc pnt pats rhs ds = HsMatch loc pnt pats (HsBody rhs) ds 
newDecl loc pnt pats rhs ds = Dec (HsFunBind loc [HsMatch loc pnt pats (HsBody rhs) ds] )

checkFree :: [PName] -> [HsDeclP] -> [HsDeclP]
checkFree [] _ = []
checkFree _ [] = []
checkFree (p:ps) list = (checkSin p list) ++ (checkFree ps list)
                                 where
                                   checkSin :: PName -> [HsDeclP] -> [HsDeclP]
                                   checkSin p [] = []
                                   checkSin p (x:xs) 
                                      | defines p x = [x]
                                      | otherwise   = checkSin p xs


{-|
Takes the position of the highlighted code and returns
the function name, the list of arguments, the expression that has been
highlighted by the user, and any where\/let clauses associated with the
function. 
-}

{-findDefNameAndExp :: Term t => [PosToken] -- ^ The token stream for the 
                                          -- file to be
                                          -- refactored.
                  -> (Int, Int) -- ^ The beginning position of the highlighting.
                  -> (Int, Int) -- ^ The end position of the highlighting.
                  -> t          -- ^ The abstract syntax tree.
                  -> (SrcLoc, PNT, FunctionPats, HsExpP, WhereDecls) -- ^ A tuple of,
                     -- (the function name, the list of arguments,
                     -- the expression highlighted, any where\/let clauses
                     -- associated with the function).
 -}                    
findDefNameAndExp toks beginPos endPos t
  = fromMaybe ([], defaultPNT, [], defaultExp, [], Def [])
              (applyTU (once_tdTU (failTU `adhocTU` inMatch `adhocTU` inPat)) t)

    where
      --The selected sub-expression is in the rhs of a match
      inMatch (match@(HsMatch loc1  pnt pats (rhs@(HsBody e)) ds)::HsMatchP)
        | locToExp beginPos endPos toks rhs /= defaultExp
          = Just ([loc1], pnt, pats, e, ds, (Match match))
      inMatch _ = Nothing

      --The selected sub-expression is in the rhs of a pattern-binding
      inPat (pat@(Dec (HsPatBind loc1 ps (rhs@(HsBody e)) ds))::HsDeclP)
        | locToExp beginPos endPos toks rhs /= defaultExp
          = if isSimplePatBind pat
              then Just ([loc1], patToPNT ps, [], e, ds, (MyPat pat))
              else error "A complex pattern binding can not be generalised!"
      inPat _ = Nothing

